"""We use our own catalog for member data"""

from zope import interface
from zope.interface import implements

from Acquisition import aq_inner, aq_parent
from AccessControl import ClassSecurityInfo, Unauthorized
from AccessControl.Permissions import manage_zcatalog_entries as ManageZCatalogEntries
from AccessControl.Permissions import search_zcatalog as SearchZCatalog

from Products.CMFCore.utils import getToolByName
from Products.Archetypes.utils import shasattr
from Products.CMFPlone.CatalogTool import CatalogTool
from Products.CMFPlone.utils import base_hasattr
from Products.CMFPlone.utils import safe_callable

from betahaus.openmember.content.memberdata import MemberData
from betahaus.openmember import CATALOG_ID, DATABASE_ID, OPENMEMBER_ROLE
from betahaus.openmember.interfaces import IOMDataHandler, IOpenMemberCatalog


class OpenMemberCatalogException(Exception):
    def __init(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class _extra:
    pass

class OpenMemberCatalog(CatalogTool):
    id = CATALOG_ID
    security = ClassSecurityInfo()
    
    implements(IOpenMemberCatalog)
    
#    def __init__(self, *args,**kwargs):
#        super(OpenMemberCatalog, self).__init__(*args,**kwargs)


    security.declareProtected(ManageZCatalogEntries, 'clearFindAndRebuild')
    def clearFindAndRebuild(self):
        """Empties catalog, then finds all contentish objects (i.e. objects
           with an omcIndexObject method), and reindexes them.
           This may take a long time.
           
           This method is overloaded to use the custom indexer method omcIndexObject instead of indexObject.
           It is also restricted to the object in site root with ID from betahaus.openmember.DATABASE_ID
        """
        def omcIndexObject(obj, path):
            if (base_hasattr(obj, 'omcIndexObject') and
                safe_callable(obj.omcIndexObject)):
                try:
                    obj.omcIndexObject()
                except TypeError:
                    # Catalogs have 'indexObject' as well, but they
                    # take different args, and will fail
                    pass
        self.manage_catalogClear()
        database = IOMDataHandler(aq_inner(self)).database
        self.ZopeFindAndApply(database, search_sub=True, apply_func=omcIndexObject)
        
    security.declareProtected(SearchZCatalog, 'searchResults')
    def searchResults(self, REQUEST=None, **kw):
        """This is a wrapper so that we can set a custom permission for searching in the catalog.
        """
        user = getToolByName(self, 'portal_membership').getAuthenticatedMember()
        if user.has_role(OPENMEMBER_ROLE):
            return super(OpenMemberCatalog, self).searchResults(REQUEST,**kw)
        raise Unauthorized('The current user does not have the '
                               'required "%s" permission' % OPENMEMBER_ROLE)
        
    __call__ = searchResults


    def isKeywordIndex(self, index):
        item = self.openmember_catalog.Indexes.get(index)
        
        if item.meta_type == 'KeywordIndex':
            return True
        return False    

    def isUIDLikeIndex(self, index):
        item = self.openmember_catalog.Indexes.get(index)
        #Reference UIDs are only stored in KeywordIndexes
        if not self.isKeywordIndex(index):
            return False
        #KeywordIndexes have items attr
        if not shasattr(item, 'items'):
            return False
        
        result = [x[0] for x in item.items()]
        if len(result):
            for val in result:
                return self.isUIDLike(val)
        
        return 'unknown'
    
    def isUIDLike(self, text):
        """ Check if something seems to be a UID.
            If it is part of a list, we can assume that something is a reference
        """
        if not text:
            return False
        if len(text) != 32:
            return False
        import re
        if re.match("[0-9a-f]{32}", text):
            return True
    
    def addIndex(self, name, index_type, label = None, extra = None):
        """Adds a index to our catalog.
           also adds a label attribute for display 
        """
        
        if index_type == 'ZCTextIndex' and extra == None:
            if not shasattr(self, 'plone_lexicon', acquire = False):
                # Make sure the lexicon is there
                pl = getattr(getToolByName(self, 'portal_catalog'), 'plone_lexicon')
                setattr(self, 'plone_lexicon', pl)

            # this is a wierd way to add the index but it is the only way I found that works.
            extra = _extra()
            extra.lexicon_id = 'plone_lexicon'
            extra.index_type = 'Okapi BM25 Rank'
            
        super(OpenMemberCatalog, self).addIndex(name, index_type, extra)
        
        # get the new index, and set the label as an attribute.
        new_index = self.Indexes.get(name)
        if label:    
            setattr(new_index, 'label', label)
        else:
            setattr(new_index, 'label', name)
        # yay we are done.
        
    def setMirrorFields(self, indexes):
        """Takes a list of ('field_name', 'IndexType')
           and adds them to the catalog as indexes and columns.
           
           Removes indexes and columns from the catalog for field_names that are not in the list.
        """
        
        omindexes = self.indexes()
        omcolumns = self.schema()

        # add indexes and columns that are not in the catalog yet.
        # prune the omindex, omcolumns so we know if there are any unneeded indexes, columns that should be removed.
        for name, meta_type, label in indexes:
            if name not in omindexes:
                self.addIndex(name, meta_type, label = label)
            else:
                this_index = self.Indexes.get(name)
                if getattr(this_index, 'label' , '') != label:
                    # Catch if the label has changed
                    this_index.label = label
                    
                if this_index.meta_type != meta_type:
                    # Catch if type is changed
                    self.delIndex(name)
                    self.addIndex(name, meta_type, label = label)
                    
                omindexes.remove(name)
                
                
            if name not in omcolumns:
                self.addColumn(name)
            else:
                omcolumns.remove(name)
                
        # remove any old indexes and columns that were not specified.
        for index in omindexes:
            self.delIndex(index)
        for column in omcolumns:
            self.delColumn(column)        
            
#    def friendlyIndexes(self, with_labels=False):
#        """Returns a list of indexes that are defined in the catalog. 
#        Indexes that are internal for the catalog and openmember are not included.
#        
#        This function is suitable for knowing what to display in user-interfaces.
#        """
#        
#        indexes = self.indexes()
#        indexes.remove('membership_references')
#        
#        if with_labels:
#            indexes = [(x,x) for x in indexes]
#        
#        return indexes