from zope.interface import implements
from zope.component import adapts

from geopy import geocoders

from Products.Archetypes.utils import shasattr
from Products.CMFCore.utils import getToolByName
#from Products.BTreeFolder2.BTreeFolder2 import manage_addBTreeFolder


from betahaus.openmember.database import manage_addOpenMemberDatabase

from betahaus.openmember import PROJECTNAME, CATALOG_ID, DATABASE_ID, logger
from betahaus.openmember.interfaces import IOMDataHandler, IOpenMemberCatalog

def onlyWhenInstalled(func):
    """A decorator that prevents ``func`` to be executed if the product isn't installed.
    """
    def wrapper(context, *args, **kwargs):
        qi = getToolByName(context, 'portal_quickinstaller')
        if qi and qi.isProductInstalled(PROJECTNAME):
            return func(context, *args, **kwargs)
        else:
            return None
    wrapper.__name__ = func.__name__
    wrapper.__dict__.update(func.__dict__)
    wrapper.__doc__ = func.__doc__
    wrapper.__module__ = func.__module__
    return wrapper

def not_member(context):
    """Objects with not_member = True should not be mirrored at all."""
    return getattr(context, 'not_member', False)
    
@onlyWhenInstalled
def added(context, event):
    """Adds member data when a content type that should be indexed is added. 
    """
    db = IOMDataHandler(context)
    db.addMember(context)
    logger.debug('IOpenMember %s was added' % context.Title())

@onlyWhenInstalled    
def modified(context, event):
    """ Modifies member data
    """
    db = IOMDataHandler(context)
    db.updateMember(context)
    logger.debug('IOpenMember %s was edited' % context.Title())

#FIXME: Not implemented yet
#def sync_profile_name(context, event):
#    """ Checks if first_name + lastname should be title and sets it.
#        This only applies for IBasePerson. (from content/mixin.py)
#    """
#    fullname = "%s %s" % (context.getFirst_name(), context.getLast_name())
#    context.setTitle(fullname)
#    context.setFullname(fullname)
#    context.reindexObject()
    
class OMDataHandler(object):
    """Handles communication with the Open Member database
    """
    implements(IOMDataHandler)
    
    def __init__(self, context, id = DATABASE_ID):
        self.site = getToolByName(context,'portal_url').getPortalObject()
        database = getattr(self.site,id,None)
        if database is None:
            logger.info('No database found\nAdding new database to portal root with id %s.' % id)
            manage_addOpenMemberDatabase(self.site, id = id, title = 'Open Member Database')
            database = getattr(self.site,id)
        
        if database.meta_type != 'BTreeFolder2':
            return None #FIXME: Throw an exception
        
        #Store for internal use.
        self.database = database
        self.catalog = getToolByName(self.site,CATALOG_ID)
        self.portal_membership = getToolByName(self.site, 'portal_membership')
        
        # maybe this should be configurable?
        self.KEY = 'ABQIAAAAQwPukS30k46hzfwPTRWMYxQ2mwsxFjngRdPU_kSKP35cSVvhVBQP-Y0USjEASHT_S0KGcW4gtvHK4w'
        self.g1 = geocoders.Google(self.KEY)
        self.g2 = geocoders.Google(resource='maps')
        
        
    def _getValueFor(self, obj, field_name):
        value = getattr(obj, 'getRaw'+field_name.capitalize(), None)
        #value might be a function (like context.Title()) rather than a property
        if callable(value):
            value = value()
        return value
        
            
    def _setAttributes(self, obj):
        """ Takes a specific object (obj) that is a regular context
            and creates/sets attributes on a MemberData object. Also takes care of versioning of data.
        """
        
        # so that geo lookup only is done once.
        has_updated_geo = False
        
        # get the properties needed for geo lookup.
        portal_properties = getToolByName(self.site, 'portal_properties', None)
        openmember_properties = getattr(portal_properties, 'openmember_properties', None)
        address_fields = []
        if openmember_properties and openmember_properties.hasProperty('address_fields'):
            address_fields = list(openmember_properties.getProperty('address_fields'))
        
        #Get the MemberData object
        memberdata = getattr(self.database, obj.UID())
        
        #Loop through all indexes to see if something has changed
        #FIXME This should be a property on the database rather than the catalog
        for index in self.catalog.indexes():
            
            # The value of the field that is to be indexed, if it exists
            # use the standard archetypes get methods
            value = self._getValueFor(obj, index)
            
            #Only set attribute if something has changed
            if value == getattr(memberdata, index, None):
                continue
            
            # this makes geo lookup depend on what is configured, 
            if not has_updated_geo and index in address_fields:
                # now we know we should do a geo lookup.
                addres_values = []
                for field in address_fields:
                    value = self._getValueFor(obj, field)
                    if value:
                        addres_values.append(value)
                
                if addres_values:
                    # get the coordinates and make sure they are set
                    coords = self._doGeoLooup(addres_values)
                    if coords != (0.0,0.0):
                        setattr(memberdata, 'geo_coords', coords)
                        has_updated_geo = True
             
            # Check if we need to write history for this property
            # Maybe this should be a property same as address_fields
            disable_history = getattr(obj, 'disable_history', [])
            if index not in disable_history:
                #History attribute name
                history = '_%s_history' % index
                
                #Get current users ID
                userid = self.portal_membership.getAuthenticatedMember().getId()
                
                #Create history attribute if it doesn't exist
                if not shasattr(memberdata, history):
                    setattr(memberdata, history, [])

                #Write current change to history
                history_obj = getattr(memberdata, history)
                history_obj.append({'data':value,
                                    'modified':memberdata.modified(),
                                    'user':userid,
                                    })
            
            #Set main value on MemberData
            setattr(memberdata, index, value)
            
    def _address2geolocation(self, address):
        """private helper function to look up a geolocation given an address
    
        address needs to be a plain (unicode) string  
    
        Raises ValueError if address cannot be resolved"""
        try:
            place, (lat, lng) = self.g2.geocode(address)
        except ValueError:
            place, (lat, lng) = self.g1.geocode(address)
            
        return (lat, lng)
            
    def _doGeoLooup(self, address_list):
        if not address_list:
            return (0.0,0.0)
        
        address = ', '.join(address_list)
        fallback1 = fallback2 = None
        if len(address_list) > 1:
            fallback1 = ', '.join(address_list[1:])
        if len(address_list) > 2:
            fallback2 = ', '.join(address_list[2:])
        
        try:
            return self._address2geolocation(address)
        except ValueError:
            if fallback1 is not None:
                try:
                    return self._address2geolocation(fallback1)
                except ValueError:
                    if fallback1 is not None:
                        try:
                            return self._address2geolocation(fallback2)
                        except ValueError:
                            pass
        
        return (0.0,0.0)
        
    def addMember(self, obj):
        """obj is the member that needs to be mapped"""
#        for index in self.indexes():
#            if prop not in self.indexes():
                # TODO: here we should not add the indexes, metadata but throw a exception if not all indexes exsist.
#                self.addIndex(prop, index_type)
#                self.addColumn(prop)
    
        #Add the actual content type
        have_memberdata = shasattr(self.database, obj.UID(), acquire = False)
        if not have_memberdata:
            self.database.manage_addProduct['betahaus.openmember'].addMemberData(obj.UID())
        else:
            return False #FIXME: Throw an exception
        
        self._setAttributes(obj)
        memberdata = getattr(self.database, obj.UID())
        self.catalog.indexObject(memberdata)
        return memberdata
                    
    def updateMember(self, obj):
        """Updates member data for obj"""
        have_memberdata = shasattr(self.database, obj.UID(), acquire = False)
        if not have_memberdata:
            return self.addMember(obj)
        else:
            self._setAttributes(obj)
            memberdata = getattr(self.database, obj.UID())
            self.catalog.reindexObject(memberdata)
            return memberdata
